/* --------------------------------------------------------------------  */
/*                          CALCULIX                                     */
/*                   - GRAPHICAL INTERFACE -                             */
/*                                                                       */
/*     A 3-dimensional pre- and post-processor for finite elements       */
/*              Copyright (C) 1996 Klaus Wittig                          */
/*                                                                       */
/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation; either version 2 of    */
/*     the License, or (at your option) any later version.               */
/*                                                                       */
/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */
/*                                                                       */
/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */
/* --------------------------------------------------------------------  */

/* ALL function prototypes for the program "CalculiX GraphiX (cgx)" */

#include "extUtil.h"
#include "trackball.h"

/* 
choose "RGB_MODE 0" for best results if you are working on a sgi workstation or
if you are working on a linux system with 8-Bit color depth. In this case 
rendering is done in the index-mode. Otherwise stay with "RGB_MODE 1"
*/
#define     RGB_MODE    1

/* gamma lightens or darkens an image (Hcpy) */
#define     GAMMA       1.9     

/* factor for the geometric tolerance for merging and radial moving */
/* the line-space will be multiplied with this value to calculate the default */
/* geometric tolerance */
#define     GTOL        1.e-5    

/* amplification factor for the automatic merging of edge-points if a cad-file is read */
#define     GTOL_EDGES  2.e1


/* defaults for calcLineDiv() */
/* defines the minimum angle between sectors of a line in calcLineDiv() */
#define GTOL_COS_A 0.998      

/* defines the default maximum distance between nodes during automatic line division calculation */
/* GTOL_NODE_DIST * gtol = target distance between nodes (not exactly achieved) */
/* A line running from one end to the other end of the model would have a division of  GTOL*GTOL_NODE_DIST */ 
#define GTOL_NODE_DIST 1.e3

/* defines the default minimum element length/maximum element length */
#define     ELEM_LENGTH_RATIO 0.1

/* default line division for meshing */
#define     DEF_LINE_DIV 4      

/* default length of one period of animation */
#define     MILLISECONDS_PER_PERIOD 1200 

/* default location of the help files, please insert the actual one */
#define     HELPFILE        {"/usr/local/CalculiX/cgx_1.3/doc/cgx/cgx.html",\
    "/usr/local/CalculiX/ccx_1.3/doc/ccx/ccx.html"} 

/* html browser, change if necessary */
/* #define     BROWSER         {"netscape"} */
#define     BROWSER         {"mozilla"}

/* postscript viewer, change if necessary */
/* #define     PSVIEWER         {"ghostview"} */
#define     PSVIEWER         {"gv"}

/* more parameters, better do not change them! */
#define     MINVALUE    -1.e9
#define     MAXVALUE     1.e9
#define     MAX_ANIM_STEPS  100
#define     INI_SCREEN      600
#define     INI_MENU_WIDTH  160
#define     INI_MENU_HEIGHT  60
#define     PS_DENSITY       107     /* basic-density to convert xwd to ps */
#define     Z_DEPTH          2.      /* edge length of the drawing cube */
#define     SET_COLS    8            /* no. of predefined colors (w,k,y,m,g,b,r) */
#define     BAS_COLS    3            /* basic colours: black, white, grey */
#define     PICK        1            /* lower nurbs-density during picking */
#define     PICK_LENGTH 5            /* default cursor-length during picking */
#define     MAX_LINE_DIV 99          /* maximum line division (cadfix restriction) */
#define     MIN_LINE_DIV 2           /* minimum line division used for the automatic division calculation */
#define     NURS_TOLERANCE 30
#define     NURS_TOLERANCE_PICK 80
#define     MAX_PARAM_PER_RECORD 20
#define     AUTOGROUP            -2  /* group number for visible elements */ 
#define     MAX_GROUPS           99999 /* maximum group number (limited by frd-format) */
#define     MAX_MATERIALS        99999 /* maximum material number (limited by frd-format) */
#define     GLUT_MENU_POINT_HEIGHT  18 /* seems to be always that value anyway(?) */
#define     GL_FEEDBACK_BUFF_SIZE 300000 /* size of the feedback buffer for nurbs-rendering */
#define     UV_STEPS        101  /* resolution of the uv-space of nurbs (triangulation ) */
#define     MIN_ANGLE_TRI3  0.999   /* badelems.c: S*S, alfa=acos(sqrt(MIN_ANGLE_TRI3)) */

/*
#define     GLUT_FONT      GLUT_BITMAP_8_BY_13
#define     GLUT_FONT      GLUT_BITMAP_9_BY_15
#define     GLUT_FONT      GLUT_BITMAP_TIMES_ROMAN_10
#define     GLUT_FONT      GLUT_BITMAP_HELVETICA_12
#define     GLUT_FONT      GLUT_BITMAP_HELVETICA_18
*/
#define     GLUT_FONT      GLUT_BITMAP_9_BY_15


#define  CMAP_CELLS_LIGHT  34             /* number of grey-values in the colormap */
#define  CMAP_DARKSHIFT     5             /* black-shift of the most bright color-cell */

/* fixed values necessary for the mesher, must not be changed */
#define  MAX_SURFS_PER_BODY   6
#define  MAX_EDGES_PER_SURF 4
#define  MAX_CORNERS_PER_BODY 8

/* for convenience fixed arrays are used in interactive functions, values can be changed */
#define  EDGES_PER_SURF 100
#define  CORNERS_PER_BODY 100
#define  SURFS_PER_BODY   100
#define  LINES_PER_EDGE  100


/* suppress the creation of nurbsobjects, only splines are made from nurb-lines */
#define NO_NURL 0
#define NO_NURS 0


/* special sets */
#define     FLPB          "-FLPB"   /* stores bodies which will be inverse oriented */ 
#define     NJBY          "-NJBY"   /*  */ 
#define     COPY          "-COPY"   /*  */
#define     MESH          "-MESH"   /*  */
#define     NOMESH        "-NOMESH" /*  */
#define     ZAP           "-ZAP"    /*  */
#define     MPC           "-MPC"    /*  */
#define     NOMPC         "-NOMPC"  /*  */
#define     NOEL          "-NOEL"   /*  */
#define     ORI           "-ORI"    /*  */
#define     DEP           "-DEP"    /* used for proj */
#define     IND           "-IND"    /* used for proj  */
#define     TMP           "-TMP"    /* temporary buffer for several purposes */
#define     HIGHLIGHT     "-HIGHL"  /* highlights identified entities */
#define     UORI          "-UORI"   /* save all unoriented items */
#define     BNUR          "-BNUR"   /* save all bad nurbs */
#define     BSUR          "-BSUR"   /* save all bad surfs */

typedef struct {
  unsigned char flpb[MAX_LINE_LENGTH];
  unsigned char njby[MAX_LINE_LENGTH];
  unsigned char copy[MAX_LINE_LENGTH];
  unsigned char mesh[MAX_LINE_LENGTH];
  unsigned char nomesh[MAX_LINE_LENGTH];
  unsigned char zap[MAX_LINE_LENGTH];
  unsigned char mpc[MAX_LINE_LENGTH];
  unsigned char nompc[MAX_LINE_LENGTH];
  unsigned char noel[MAX_LINE_LENGTH];
  unsigned char ori[MAX_LINE_LENGTH];
  unsigned char dep[MAX_LINE_LENGTH];
  unsigned char ind[MAX_LINE_LENGTH];
  unsigned char tmp[MAX_LINE_LENGTH];
  unsigned char highl[MAX_LINE_LENGTH];
  unsigned char uori[MAX_LINE_LENGTH];
  unsigned char bnur[MAX_LINE_LENGTH];
  unsigned char bsur[MAX_LINE_LENGTH];
} SpecialSet;

typedef struct {
  float r, g, b;
} Colours;


typedef struct {
  unsigned char n, e, p, l, c, s, b, L, S, se;
} Lchar;


typedef struct {
    float r;
    int i;
}Rsort;


typedef struct {
    char typ;
    int indx;
    int nr;
}CfdSurf;

int improveBadTr3(unsigned char *setname);

/* writefbd.c */
int writefbd(unsigned char *setname, SumGeo *anz, Points *pnt, Lines *line, Lcmb *lcmb, Gsur *surf, Gbod *body, Nurbs *nurs, Sets *set );

/* XFunktions_mesa.h */
void calcOffset(void);
void getColormap(void);
void storeColors(int anz_col, int pix );

/* copyEntity.h */
int createCenterpnt( unsigned char *pkt1, unsigned char *pkt2, unsigned char *pkt3 );
int copy_set( int settrgt, unsigned char *trans, int setNr, int *dep_p, int *dep_l, int *dep_c, int *dep_s, int *dep_se);
void pre_swep( unsigned char *record );
void pre_copy( unsigned char *record );
int  body_( unsigned char *name, unsigned char *corner );

/* dataGeo.h */
void descalPoints ( int anz_p, Points *point, Scale *scale );
void scalPoints ( int anz_p,  Points *point, Scale *scale );
void descalSurfs ( int anz_s,  Gsur *surf, Scale *scale );
void scalSurfs ( int anz_s,  Gsur *surf, Scale *scale );
void getScaleValues( int anz_n, int anz_p, Points *point, Nodes *node, Scale *scale);

/* dataMesh */
void descalNodes ( int anz_n, Nodes *node, Scale *scale );
void scalNodes ( int anz_n, Nodes *node, Scale *scale );
void calcDatasets( int num_olc, Summen *anz, Nodes *node, Elements *elem,
		    Datasets *lcase );
void calcAnimation( GLuint list_animate_light, GLuint *list_animate, int anim_steps, float anim_faktor,
                    int *anim_alfa, int halfperiode, int centerNode, Nodes *n_anim, int  elemGroupNr, Normalen *norm,
                    int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3,
                    CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CPenta20 *cpenta20, int lc,
                    Summen *anz, Scale *scale, Datasets *lcase,  CHexa20 *chexa20, CHexa8 *cquad8 );
void calcSequence( GLuint list_animate_light, GLuint *list_animate, int *seqLC, int seq_steps, float anim_faktor, int halfperiode, int centerNode, Nodes *n_anim, int  elemGroupNr, Normalen *norm, int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3, CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CPenta20 *cpenta20, int lc, Summen *anz, Scale *scale, Datasets *lcase,  CHexa20 *chexa20, CHexa8 *cquad8 );

/* defineEntity.h */
int hashAlias( SumAsci *sumAsci, unsigned char *name, int nr);
void operateAlias( unsigned char *name, unsigned char *type );
void delNod( int anzn, int *number );
int  nod( Summen *anz, Nodes **nptr, int flag, int nodnr, float x, float y, float z, int scalFlag );
void pre_nod( unsigned char *record);
void delElem( int anze, int *index );
int elem_CBeam( int elnr, int n1, int n2, int setFlag, int eattr );
int elem_CBeam3( int elnr, int n1, int n2, int n3, int setFlag, int eattr );
int elem_CTri3( int elnr, int n1, int n2, int n3, int setFlag, int eattr );
int elem_CTri6( int elnr, int n1, int n2, int n3, int n4, int n5, int n6, int setFlag, int eattr );
int elem_CQuad4( int elnr, int n1, int n2, int n3, int n4, int setFlag, int eattr );
int elem_CQuad8( int elnr, int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8, int setFlag, int eattr );
int elem_CHexa8( int elnr, int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8, int setFlag, int eattr  );
int elem_CTet4( int elnr, int n1, int n2, int n3, int n4, int setFlag, int eattr );
int elem_CHexa20( int elnr, int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8, int n9, int n10, int n11, int n12, int n13, int n14, int n15, int n16, int n17, int n18, int n19, int n20, int setFlag, int eattr );
int elem_CTet10( int elnr, int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8,
		 int n9, int n10, int setFlag, int eattr );
int hashPoint( SumAsci *sumAsci, unsigned char *name, int nr);
int hashLine( SumAsci *sumAsci, unsigned char *name, int nr);
int hashLcmb( SumAsci *sumAsci, unsigned char *name, int nr);
int hashSurf( SumAsci *sumAsci, unsigned char *name, int nr);
int hashBody( SumAsci *sumAsci, unsigned char *name, int nr);
int hashSet( SumAsci *sumAsci, unsigned char *name, int nr);
void pre_elem( unsigned char *record);
void bak_delPnt( int anz, int *number );
void delPnt( int anz, int *number );
int  pnt( unsigned char *name, float x, float y, float z, int scalFlag );
void pre_pnt( unsigned char *record, int flag);
void delLine( int anz, int *number );
int line_i( unsigned char *name, int ip1, int ip2, int trk, int div, float bias, unsigned char type );
int  line_( unsigned char *name, unsigned char *p1, unsigned char *p2, unsigned char *trk, int div, float bias );
void pre_line( unsigned char *record, int addFlag);
void delLcmb( int anz, int *number );
int lcmb_i( unsigned char *name, int add, int anz_l, unsigned char  *ori, int *lin );
int  lcmb_( unsigned char *name, int add, int anz_l, unsigned char *ori, unsigned char *lin );
void pre_lcmb( unsigned char *record, int addFlag);
void bak_delSurf( int anz, int *number );
void delSurf( int anz, int *number );
int  surface_i( unsigned char *name, unsigned char ori, int blend, int anz_c, unsigned char *cori, int *edge, unsigned char *typ );
int  surface( unsigned char *name, unsigned char ori, unsigned char *blend, int anz_c, unsigned char *cori, unsigned char *corner );
void pre_gsur( unsigned char *record, int addFlag);
void pre_surf( unsigned char *record);
void bak_delBody( int anz, int *number );
void delBody( int anz, int *number );
int  gbod_i( unsigned char *name, int blend, int anz_c, unsigned char *cori, int *corner );
int  gbod( unsigned char *name, unsigned char *blend, int anz_c, unsigned char *cori, unsigned char *corner );
void pre_gbod( unsigned char *record, int addFlag);
void pre_body( unsigned char *record);
int  nurl( unsigned char *string, int addFlag );
void delNurs( int anz, int *number );
int hashNurs( SumAsci *sumAsci, unsigned char *name, int nr);
int  nurs( unsigned char *string, int addFlag );


/* dispLists.h */
void drawSurfaces_vector( int lc, int ne, int *e, float v_factor);
void drawElements_vector( int , int , int *, float);
void getCTri3Normalen( CTri3 *ctri3, Nodes *node, int *num_etype, Normalen *norm );
void getCTri6Normalen( CTri6 *ctri6, Nodes *node, int *num_etype, Normalen *norm );
void getCQuad4Normalen( CQuad4 *cquad4, Nodes *node, int *num_etype, Normalen *norm );
void getCQuad8Normalen( CHexa8 *cquad8, Nodes *node, int *num_etype,  Normalen *norm);
void getCHexa8Normalen( CHexa8 *chexa8, Nodes *node, int *num_etype, Normalen *norm );
void getCPenta6Normalen( CPenta6 *cpenta6, Nodes *node, int *num_etype, Normalen *norm );
void getCTet4Normalen( CTet4 *ctet4, Nodes *node, int *num_etype,  Normalen *norm);
void getCTet10Normalen( CTet10 *ctet10, Nodes *node, int *num_etype,  Normalen *norm);
void getCHexa26Normalen( CHexa20 *chexa20, Nodes *node, int *num_etype,  Normalen *norm);
void drawModelEdges( GLuint list_model_edges, int color, float width,
                     int numEdges, Nodes *node, Edges *edge );
void drawSurfEdges( GLuint list_surf_edges, int color, float width, int  elemGroupNr,
                     int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CHexa8 *surf8,
		    CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTri3 *surf3, CTri6 *surf6, CHexa8 *cquad8 );
void drawElemEdges( GLuint list_elem_edges, int color, float width, int  elemGroupNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
		    CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTet4 *ctet4, CTet10 *ctet10, CHexa8 *cquad8, CPenta6 *cpenta6, CPenta20 *cpenta20 );
void drawSurfaces_light( GLuint list_surf_light, unsigned char cullFlag, int  elemGroupNr, Normalen *norm,
                     int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3,
			 CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CHexa8 *cquad8 );
void drawElements_light( GLuint list_elem_light,  unsigned char cullFlag, int  elemGroupNr,  Normalen *norm,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
			 CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTet4 *ctet4, CTet10 *ctet10, CHexa8 *cquad8, CPenta6 *cpenta6, CPenta20 *cpenta20 );
void drawSurfaces_load( GLuint list_surf_load,  unsigned char cullFlag, int  elemGroupNr, int *colNr,
                     int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3,
			CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CHexa8 *cquad8 );
void drawElements_load( GLuint list_elem_load,  unsigned char cullFlag, int  elemGroupNr, int *colNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
			CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTet4 *ctet4, CTet10 *ctet10, CHexa8 *cquad8, CPenta6 *cpenta6, CPenta20 *cpenta20 );
void drawElements_elstress( GLuint list_elem_elstress,  unsigned char cullFlag, int  elemGroupNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
			    CTri3 *ctri3, CQuad4 *cquad4, CBeam *cbeam );
void drawSurfaces_elstress( GLuint list_elem_elstress,  unsigned char cullFlag, int  elemGroupNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
                     CTri3 *ctri3, CQuad4 *cquad4, CBeam *cbeam );


/* dispLists_rgb.c  rgb mode */
void drawSurfaces_load_rgb( GLuint list_surf_load,  unsigned char cullFlag, int  elemGroupNr, Colours *colRgb,
                     int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3,
			CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CHexa8 *cquad8 );
void drawElements_load_rgb( GLuint list_elem_load,  unsigned char cullFlag, int  elemGroupNr, Colours *colRgb,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
			CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTet4 *ctet4, CTet10 *ctet10, CHexa8 *cquad8, CPenta6 *cpenta6, CPenta20 *cpenta20 );
void drawElements_elstress_rgb( GLuint list_elem_elstress,  unsigned char cullFlag, int  elemGroupNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
			    CTri3 *ctri3, CQuad4 *cquad4, CBeam *cbeam );
void drawSurfaces_elstress_rgb( GLuint list_elem_elstress,  unsigned char cullFlag, int  elemGroupNr,
                     int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8,
                     CTri3 *ctri3, CQuad4 *cquad4, CBeam *cbeam );


/*  extFunktions.h */
void help( void );
int compareInt(int *a, int *b);
int compareFloat(float *a, float *b);
int askGLError(unsigned char buffer[134]);
void defineColIndexes_light(void);
void defineColIndexes_load(void);
void defineColIndexes_set(void);
void zoom(float xcur0, float ycur0, float xcur1, float ycur1);
void qzoom(void);
void center(float x, float y, float z);
void qcenter(int x, int y);
void rot_u(float a);
void rot_r(float a);
void rot_c(float a);
void rot_x(float a);
void rot_y(float a);
void rot_z(float a);
void pre_rot( unsigned char *record);
int transform( unsigned char *record, int anz_n, Nodes *node );
int calcLineDiv(Lines *line, int nr, Points *point, float gtol_cos_a, float lmax, float lmin);
int fillBlendedSurf(int nr);
void repNurl(int nr );
void repNurs(int nr );
void untrimNurs(int nr );
void repSurf(int nr );
void rep( int j );
void adjustNodes();


/* cgx.h */
void nodalDataset( int auswahl, int lc, Summen *anz, Scale *scale, Nodes *node,
                    Datasets *lcase, float *value_n, int *num_etype, CHexa20 *chexa20, CPenta20 *cpenta20, Colours *colRgb
                   , int *colNr, int offset, int maxIndex, int steps );
void elementDataset( int auswahl, int lc, Summen *anz, Scale *scale, Datasets *lcase,
                      int *num_etype, CBeam *cbeam, int offset, int maxIndex, int steps );
void allocUtilityData(void);
void entryfunktion( int state );
void WindowState( int state );
void Mouse( int x, int y );
void MouseState( int button, int state, int x, int y );
void menu( int selection );
void updateDispLists(void);
void ConfigureAndShowWindow_Light(void);
void ConfigureAndShowWindow_Load( void );
void ConfigureAndShowWindow_Sequence( void );
void ConfigureAndShowWindow_Vector( void );
int minus( unsigned char *record );
int  plot( unsigned char *record );
int  plus( unsigned char *record );
void changeAnimation( int selection );
void newTimePerPeriode( int selection );
void tuneAnimation( int selection );
void stepsAnimation( int selection );
void redraw(void);
void selectEntity( int selection );
void createDatasetEntries(void);
void selectDataset( int selection );
Nodes *addDispToCoordinates( Nodes *nodes);
void selectView( int selection );
void selectData( unsigned char *record);
void Keyboard( unsigned char gkey, int x, int y );
void moveModel(void);
void DrawGraficLoad( void );
void DrawGraficLight( void );
void DrawGraficAnimate( void );
void DrawGraficSequence( void );
void DrawGraficVector( void );
void DrawPickedItems(void);
void DrawAxes();
void iniDrawMenu(void);
void DrawMenuLoad( void );
void DrawMenuLight( void );
void DrawMenuAnimate( void );
void DrawMenuSet( void );
void reshape( int width, int height );
void initLightAndMaterial_rgb( void );
void initLightAndMaterial_index( void );

/* iniMeshData.h     */
void makeSurfaces(void);
void iniMeshData( unsigned char *datin, unsigned char *format );
void iniElements(Summen *anz, Elements *elem, int ***facebuf);

/* mergEntity.h */
int *findCloseEntity( unsigned char type, Nodes *node, Points *point, Rsort *rsort, int indx, int anz_n, float local_gtol);
int mergeNode( Rsort *rsort, int anz_n, int indx, float local_gtol, int lock );
int mergePnt( Rsort *rsort, int anz_p, int indx, float local_gtol, int lock );
int mergeLine( int indx, int lock );
int mergeLcmb( int indx );
int mergeSurf( int indx, float local_gtol );
int compareRsort(Rsort *a, Rsort *b);
void pre_merge( unsigned char *record );

/* orient.h */
int orientLcmb( int nr );
int orientSurf( int nr );
int ori7SidedBody( int nr, int **cp );
int ori6SidedBody( int nr, int **cp );
int ori5SidedBody( int nr, int **cp );
int orientBody( int nr );
void orientSet( unsigned char *record);

/* pickFunktions.h    */
int hitAction( GLuint *name, unsigned char *type, int x, int y );
int hitUndo( GLuint *name, unsigned char *type, int x, int y );
int pickstack( GLuint name );
int processHits( GLint hits, GLuint *buffer, unsigned char *type, unsigned char *mode, int x, int y );
void defineDiv( unsigned char gkey, int x, int y );
void pick( unsigned char gkey, int x, int y );
void goPicking( int x, int y, unsigned char *type );
void drawListsPick(void);
void qali(void);
void qbia(void);
void qnod(void);
void qcnt(void);
void qdis(void);
void qele(void);
void qenq(void);
void qflp(void);
void qdiv(void);
void qadd( unsigned char *record);
void qdel( void );
void qrem( unsigned char *record);
void qpnt( unsigned char *record);
void qlin( unsigned char *record);
void qsur( unsigned char *record);
void qbod( unsigned char *record);
void qseq(void);
void qspl(void);
void moveNode(int n, int x, int y);
void movePoint(int p, int x, int y);
int createElem(int n);
int createPoint( int x, int y );
int createLine( unsigned char *pnt, int flag );
void convertLine( int line, int pickbuf );
int splitLine( int lin, int x, int y  );
void pre_align( int nr, int flag );
void drawListsPick(void);
void moveModel(void);
void updateDispLists(void);
void flip( unsigned char *type, int indx);

/* sendSet.h  */
void sendSurfNormalen( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, Datasets *lcase , Scale *scale );
void sendPressure( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *value, unsigned char *val1 );
void sendFilm( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *val1, unsigned char *val2, unsigned char *val3, unsigned char *val4 );
void sendRadiate( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *val1, unsigned char *val2, unsigned char *val3, unsigned char *val4 );
void sendCflux( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *value, unsigned char *val2 );
void sendDflux( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *val1, unsigned char *val2 );
void sendSPC( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *value, unsigned char *val2 );
void sendSPC( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *value, unsigned char *val2 );
void sendSliders( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *spcType );
void sendNames( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire );
void sendForce( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, float *f );

/* setFunktions.h  */
void iniSetFunktions(void);
void drawSurfNodes_plot( Summen *anz, int  elemGroupNr, int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3, CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CHexa8 *cquad8 );
void drawSurfaces_plot( unsigned char cullFlag, int  elemGroupNr, int *num_stype, int *num_etype, Nodes *node, CQuad4 *surf4, CTri3 *surf3, CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CHexa8 *cquad8 );
void drawElemNodes_plot( Summen *anz, int  elemGroupNr, int *num_etype, Nodes *node, CHexa20 *chexa20, CHexa8 *chexa8, CTri3 *ctri3, CTri6 *ctri6, CQuad4 *cquad4, CBeam *cbeam, CBeam3 *cbeam3, CTet10 *ctet10, CHexa8 *cquad8  );
void drawNodes_plot( int num, int *name, Nodes *node , int col, unsigned char type);
void drawElements_plot( int num, int *name, Nodes *node, Elements *e_enqire, int col, unsigned char type );
void drawFaces_plot( int num, int *name, Nodes *node, Faces *face, int col, unsigned char type );
void drawPoints_plot( int num, int *name, Points *point , int col, unsigned char type);
void drawLines_plot( int num, int *name, Lines *line , Points *point, int col, unsigned char type);
void drawSurfs_plot( int num, int *name, Gsur *surf, Lcmb *lcmb, Lines *line , Points *point, int col, unsigned char type);
void drawBodys_plot( int num, int *name, Gbod *body, Gsur *surf, Lcmb *lcmb, Lines *line , Points *point, int col, unsigned char type);
void drawNurl_plot( int num, int *name, int col, unsigned char type, int pickflag);
void drawNurs_plot( int num, int *name, int col, unsigned char type, int pickflag);
int getSetNr(unsigned char *setname);
int getAliasNr(unsigned char *name);
int getNodNr(Summen *anz,Nodes *node,int nodnr );
int getElemNr (int elnr );
int getE_enqNr (int elnr );
int getPntNr(unsigned char *name);
int getLineNr(unsigned char *name);
int getLcmbNr(unsigned char *name);
int getSurfNr(unsigned char *name);
int getBodyNr(unsigned char *name);
int getNurlNr(unsigned char *name);
int getNursNr(unsigned char *name);
void delSet( unsigned char *setname);
int seto( unsigned char *setname );
void setc( unsigned char *setname );
int pre_seta( unsigned char *setname, unsigned char *type, unsigned char *name);
int seta( int setNr, unsigned char *type, int number);
int setr( int setNr, unsigned char *type, int number);
void seqr( int setNr, int number);
void pre_del( unsigned char *record);
void pre_move( unsigned char *record);
void free_proj(float *nx, float *nrad, int *nmissed  );
void pre_proj( unsigned char *record);
int prnt(unsigned char *record);
int  zap(unsigned char *record);
void grpa( int *elemGroup, int nr, int setNr );
void mata( int *elemMat, int nr, int setNr );
unsigned char getLetter(int i);
int getFamName( int n, unsigned char *c );
int getNewName( unsigned char *name, unsigned char *type );
void pre_movie(unsigned char *string);
int commandoInterpreter( unsigned char *type, unsigned char *string, int na, int nb, FILE *handle1, int addFlag );
void readfbd( unsigned char *datin, int addFlag );
void readfrdfile( unsigned char *datin, unsigned char *setname );
void pre_read( unsigned char *record );
void pre_elty( unsigned char *record );
void pre_mesh( unsigned char *record );
int completeSet( unsigned char *setname, unsigned char *type );
int sendTriangles(  unsigned char *setname );
int write2stl(unsigned char *datout, Summen *anz, Nodes *node, Elements *elem, Datasets *lcase );
int write2netgenEdges(unsigned char *datout, Summen *anz, Nodes *node, Elements *elem, Datasets *lcase );
void pre_write( unsigned char *record );
void senddata( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire,
               Datasets *lcase , Scale *scale, int complete, int lc );
void drawSets(int mode);

/* sendMpc.c */
void sendMpc( unsigned char *setname, unsigned char *format, Summen *anz, Nodes *node, Elements *e_enqire, unsigned char *rotation, float *vector  );
void cycmpc(int set1, int set2, unsigned char *format, unsigned char *value, unsigned char *corr);
void areampc(int set1, int set2, unsigned char *format, unsigned char *value, unsigned char *corr);
int calcCoefficients(  Nodes *node, int n0, CTri3 *ctri3, int  e, double *c);
float find_Tri3( Nodes *node, int n, CTri3 *ctri3, int e);
void corrNode(  Nodes *node, int n, CTri3 *ctri3, int  e, float g );
void createMPCs();

/* graph.c */
void plot2D(unsigned char type, int setNr );
void graph( unsigned char *record);

/* corrad.c */
void corrad( int setNr, float minradius);

/* badelems.c */
int calcBadElements(unsigned char *record);

/* write2tochnog.c */
int write2tochnog(unsigned char *datout, Summen *anz, Nodes *node, Elements *elem, Datasets *lcase );

/* readccx.c */
int readccx(unsigned char *datin, Summen *anz, Nodes **nptr, Elements **eptr, Datasets **lptr, int ****fptr );
int getFilePointer( int filesopen, FILE **openfile, unsigned char *rec_str);


double repairNurbs( Nurbs *nurbs, int nr, int deg, int dir);
int evalNurbs( Nurbs *nurbs, int nr, int sum_p, double *pnt_u, double *pnt_v, Points *pnt);
void projSurfToNurbs( Nurbs *nurbs, int nr, Gsur *surf, int snr, Nodes *node );
void projPntsToNurbs( Nurbs *nurbs, int nr, Sets *set, int setNr, Points *pnt);
void trimNurbs( Nurbs *nurbs, int nr, int patch, double);


/* fillBody2.c */
int  fillSurf_ab( int j, int umax, int vmax, int *n_uv, int amax, int bmax, int *n_ab, int offs1, int offs2, float *x, float *y, float *z );
int  fillSurf_ba2( int sur, int *div_l, int div_a, int div_b, int sa1, int sa2, int sb1, int sb2, int *n_uv, int *umax, int *vmax, int *n_ba, int *amax, int *bmax,  int *offs_sa1, int *offs_sa2, float *x, float *y, float *z, float *x_ba, float *y_ba, float *z_ba );
int fillBody2( int b_indx, int *srefp, int **div_l, int *unbalance, int **nabc, int *amax, int *bmax, int *cmax, int *offs_sa1, int *offs_sa2, int *msur, int *div_sa1, int *div_sa2 );


/* meshSet.c */
void cornerNodes( int vmax, int umax, int j, int *n_uv );
int  newCornerPositions( int n, int u,int v,int umax, int vmax, float *lx,float *ly,float *lz,
 float *x,float *y,float *z, int flag );
int  fillSurf( int j, int *n_uv, int umax, int vmax, float *x, float *y, float *z );
int  fillSurf2( int sur, int *div_l, int div_a, int div_b, int sa1, int sa2, int sb1, int sb2,
 int *n_uv, int *umax, int *vmax, int *n_ba, int *amax, int *bmax, int *offs_sa1, int *offs_sa2,
 float *x, float *y, float *z );
void edgeNodes( int vmax, int umax, int j, int *n_uv );
int  newEdgePositions( int n, int u,int v,int umax, int vmax, float *lx,float *ly,float *lz, float *x,float *y,float *z, int flag );

int meshPoints( int setNr );
int straightNodes( int j, int k, int d, float *pn );
int arcNodes( int j, int k, int d, float *pn );
int splineNodes( int j, int k, int d, float *pn );
int nurlNodes( int j, int k, int d, float *pn );
int meshLines( int setNr );
int newDivisions( int *div_l, int *div_a,int *div_b, int *sa1,int *sa2,int *sb1,int *sb2 );
int addTwoLines( int l1, unsigned char o1, unsigned char typ1, int l2, unsigned char o2, unsigned char typ2 );
int splitLineAtDivratio(int edge, int typ, float splitdiv, int *lnew, unsigned char  *typnew);
int mesh_tr3u(int nr );
int merge_nlocal( Rsort *rsort, Nodes *nlocal, CTri3 *ctri3, int anz_e, int anz_n, int indx, int lock, float tol );
int deleteFreeEdges(Nodes *node, int *snods, int anz_n, CTri3 *elems, int anz_e );
int mesh_tr3g( int s );
int meshSurfs( int setNr );
int fillBody( int b, int *srefp, int umax, int vmax, int wmax, int *n_uvw, float *x, float *y, float *z );
int getSurfDivs( int b, int s, int *div_l );
int meshImprover( int *etyp, int *nodes, int *elements, int *n_indx, int *n_ori, float **n_coord, int **e_nod, int *n_type, int **n_corner, int **s_div );
int bodyFrom5Surfs( int *b);
int bodyFrom7Surfs( int *b);
int meshBodys( int setNr );
void meshSet(unsigned char *setname);


/* surfMesh.h */
int surfMesh(int *imax, int *jmax, float *x, float *y, float *z);

/*  bodyMesh2.h */
int bodyMesh2(int *wmax, int *vmax, int *umax, float *x, float *y, float * z);

/*  improveMesh.o */
int smooth_( double *co, int *nk, int *kon, int *ne, int *iptr1, int *iptr2, int *nside, int *ineino, int *nkind, int *neigh, int *ndiv, double *pneigh, int *maxsumdiv, int *etyp);

/* userFunction.c */
void userFunction(unsigned char *string, Summen *sum, Nodes **nptr, Elements **eptr, Datasets **lptr );

/* calcWeight.c */
#define real float
#define integer int

/* Subroutine */ int hexaeder_(integer *iel, integer *istat, real *x, real *y,
			       real *z__, real *v, real *xcg, real *ycg, real *zcg);
/* Subroutine */ int tetraeder_(integer *iel, integer *istat, real *x, real *y,
				real *z__, real *v, real *xcg, real *ycg, real *zcg);
