// libSNL - Simple Nurbs Library
// Copyright Scott A.E. Lanham, Australia.
// ---------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU Library General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// *** General NURBS Surface ***

#ifndef SNLSURFACE_H
#define SNLSURFACE_H

#include "snlPoint.h"
#include "snlCtrlPointNetSurface.h"
#include "snlKnotVector.h"
#include "snlVertex.h"
#include "snlVertexNet.h"
#include "snlMeshable.h"
#include "snlCurve.h"

#include "ptrList.h"

#define SNL_INV_ITER_VP 8  // Number of iterations for velocity pass.
#define SNL_INV_ITER_NP 8  // Number of iterations for newton pass.

#define SNL_NUMERIC_NOISE 1.0e-14

typedef struct
{
    /* Surface Edge */

    int     direction;  // u = 0, v = 1.
    knot    pVal;  // Parameter value.

} sEdge;

typedef struct
{
    // Surface convex hull section data item.

    snlVector   insideNorm;  // Normal pointing to inside of hull.
    int         locusIndex;  // Index in control point array of locus point.
    
} snlHullItem;

typedef struct
{
    // surface location.

    knot        paramU;  // Evaluated parameter in u direction. Point found.
    knot        paramV;  // Evaluated parameter in v direction.
    snlPoint    pt;  // Point corresponding to parameters.
    basis       dist;  // Distance from point being inverted to point found.
    double      cos;  // Cos of angle to normal.
    
    int         origPtIndex;  // For inversion and projection, the index of the original point handed to the
                              // relevant function.
                              
} snlSurfLocn;

typedef struct
{
    // surface location.

    knot        paramU;  // Evaluated parameter in u direction. Point found.
    knot        paramV;  // Evaluated parameter in v direction.
    snlPoint    pt;  // Point corresponding to parameters.
    basis       dist;  // Distance squared from point being inverted to point found.
    double      cos;  // Cos of angle to normal.
    
    int         origPtIndex;  // For inversion and projection, the index of the original point handed to the
                              // relevant function.

    int         spanNumber;  // Absolute span number guess belongs to.

    knot        minU;  // Parametric boundaries used for culling.
    knot        maxU;
    knot        minV;
    knot        maxV;

    bool        culled;
    bool        ignoreParamBounds;  // Don't cull this point if it goes out of bounds.
    bool        converged;
    
} snlSurfLocnGuess;

typedef struct
{
    // Control point location for surface

    int             uIndex;  // Array indexes.
    int             vIndex;

    snlCtrlPoint*   ctrlPt;  // Pointer to control point found.

    double          dist;  // Distance to control point from comparison point.
    
} snlSCtrlPtLocn;

typedef struct
{
    bool        converged;
    bool        stalled;  // Can't converge

    int         arcIndex;

    double      arcAngle;

    // Controlling Surface.

    knot        cParamU;
    knot        cParamV;

    snlPoint    cPt;  // Evaluated point.
    snlPoint    normPt;  // Normal point.

    snlVector   velU;  // U direction velocity.
    snlVector   velV;  // V direction velocity.

    // Matching Surface.

    knot        mParamU;
    knot        mParamV;

    snlPoint    mPt;  // Evaluated point.

} arcLocn;

class snlSurface : public snlMeshable
{
    public:
        
        snlSurface();
        virtual ~snlSurface();
    
        snlSurface ( int degU, int degV, unsigned sizeU, unsigned sizeV,
                     snlPoint& origin, snlPoint& cornerMaxU, snlPoint& cornerMaxV );
                     
        snlSurface ( int degU, int degV, unsigned sizeU, unsigned sizeV, snlCtrlPoint* points,
                     knot* knotsU, knot* knotsV );
        
        snlSurface ( const snlSurface& );  // Copy constructor.
        
        snlSurface ( snlCurve& curve1, snlCurve& curve2 );  // Generate ruled surface.

        snlSurface ( snlCurve& generator, snlPoint& axisStart, snlPoint& axisEnd, double angle );  // Surface of revolution.

        snlSurface ( snlCurve** curves, int numCurves, int dir = SNL_U_DIR );  // Skinned surface.

        // Interpolation Functions.

        enum SNL_INTERP_TYPES
        {
            SNL_GLOBAL_INTER_CHORDLENGTH,
            SNL_GLOBAL_INTERP_CENTRIFUGAL
        };

        knot* globalInterpGenParams ( int type, snlPoint* points, int sizeU, int sizeV, int dir );
        
        // Data functions.
        
        int degreeU() const;
        int degreeV() const;
        
        unsigned sizeU() const;
        unsigned sizeV() const;

        knot minU();
        knot maxU();
        
        knot minV();
        knot maxV();
        
        const snlCtrlPoint* controlPoints();
        
        const knot* knotsU();
        const knot* knotsV();
        
        snlCtrlPointNetSurface& controlPointNet();
        
        const snlKnotVector& knotVectorU();
        const snlKnotVector& knotVectorV();
        
        // Evaluation functions.

        // Non-rational homogeneous surface point.        
        snlPoint evalHmg ( knot paramU, knot paramV, basis* basisU = 0, basis* basisV = 0 ) const;
        
        // Rational non-homogeneous surface point.
        virtual snlPoint eval ( knot paramU, knot paramV, basis* basisU, basis* basisV ) const;
        virtual snlPoint eval ( knot paramU, knot paramV ) const;

        // Derivatives.

        snlPoint* evalDerivsHmg ( knot paramU, knot paramV, unsigned derivU, unsigned derivV,
                                  basis* basisU = 0, basis* basisV = 0 );
                                  
        snlPoint* evalDerivs ( knot paramU, knot paramV, unsigned derivU, unsigned derivV );

        void velocities ( knot paramU, knot paramV, snlPoint& evalPoint, snlVector& velocityU, snlVector& velocityV,
                          basis* basisU = 0, basis* basisV = 0 );
        
        void insertKnot ( knot iParam, int dir, bool reallocate = true );
        void insertKnot ( knot iParam, int dir, int numToInsert, bool reallocate = true );
        double removeKnot ( unsigned removalIndex, int direction, double tolerance, bool reallocate = true );
        
        snlVertex* project_depr ( snlPoint* toProject, int numPoints, double convergTol, double normTol, int maxPass );

        snlSurfLocn* invert ( snlPoint* toInvert, int numPoints, int* retArraySize,
                              double convergTol, double normTol, int maxPass );

        snlSurfLocn* project ( snlPoint* toProject, int numPoints, int* retArraySize,
                               double convergTol, double normTol, int maxPass );

        snlSurfLocn* fastProject ( snlPoint* toProject, int numPoints, int* retArraySize,
                                   double convergTol, double normTol, int maxPass,
                                   int sensitivity, int maxLocns );

        snlSurfLocn* processGuesses ( snlPoint* points, int numPoints, int* retArraySize,
                                      ptrList <snlSurfLocnGuess>* guesses, double convergTol,
                                      double normTol, int maxPass, bool retNonConverged = false, bool noCull = false,
                                      int numVelocity = SNL_INV_ITER_VP, int numNewton = SNL_INV_ITER_NP );

        bool convergeVelocity ( snlPoint* convergToPts, ptrList <snlSurfLocnGuess>* guesses,
                                int numIterations, double convergTol, double normTol );

        bool convergeNewton ( snlPoint* convergToPts, ptrList <snlSurfLocnGuess>* guesses,
                              int numIterations, double convergTol, double normTol );

        ptrList <snlSurfLocnGuess>* guessInvLocation ( snlPoint* points, int numPoints, bool* pointMask,
                                                       int granU, int granV );

        ptrList <snlSurfLocnGuess>* guessProjLocation ( snlPoint* points, int numPoints, bool* pointMask );

        ptrList <snlSurfLocnGuess>* guessFastProjLocation ( snlPoint* points, int numPoints, int maxGuessPerPt,
                                                            int granU, int granV );
        
        snlSCtrlPtLocn* findClosestCtrlPt ( snlPoint* points, int numPoints );
        
        snlHullItem* genHullData ( int uStart, int uEnd, int vStart, int vEnd );
        
        // Try to predict edges that may have ambiguities during projection.
        int hasAmbigEdges ( sEdge* results, double tolerance );
        int hasAmbigEdges_depr ( sEdge* results );
        
        unsigned createBezierSegments ( int dir );
        void createBezierSegments ( int* numU = 0, int* numV = 0 );
        void createConvexBezierSegments ( int* numU = 0, int* numV = 0, double sensitivity = 0.0 );
        
        double reduceDegree ( int dir, unsigned numDeg, double tolerance );
        
        void refine ( double tolerance );        
        void refineU ( double tolerance );
        void refineV ( double tolerance );
        
        double maxCurvatureU();
        double maxCurvatureV();

        snlVector calcNormal ( knot paramU, knot paramV, snlPoint* evalPt = 0 );

        snlCurve* extractEdge ( int edge );

        snlSurface* fillet ( int edge, snlVector& frontFaceNormal,
                             snlSurface& surface2, snlVector& frontFaceNormal2,
                             double tolerance, double radius, bool trim1, bool trim2 );

        void transform ( snlTransform& transf );
        
        void print();

        void print_cpp();
        
        // Abstract Implementation.
        
        virtual void vertexNet ( snlVertexNet* vNet, double tolerance, bool parametric );
        
        enum parametricDirections
        {
            SNL_U_DIR = 0,
            SNL_V_DIR = 1
        };

        enum surfaceEdges
        {
            SNL_EDGE_UMIN,
            SNL_EDGE_UMAX,
            SNL_EDGE_VMIN,
            SNL_EDGE_VMAX
        };
        
    protected:
    
        int     degU;  // Degree of surface in U direction.
        int     degV;  // Degree of surface in V direction.
    
        snlCtrlPointNetSurface*     ctrlPtNet;
        
        snlKnotVector*              knotVectU;
        snlKnotVector*              knotVectV;
        
};

#endif
